{-# OPTIONS_GHC -fth -fglasgow-exts #-}

-- Tom Moertel <tom@moertel.com>
-- 2006-03-24

module Text.Safe.Strings.Core
(
  Language(..)
, SafeString
, empty, frag, text, polytext
, cat, (+++)
, isEmpty
, render, renders, lang
)
where

-- | The Language class represents string-like types that are subject
-- to interpretation, e.g., XML or SQL.  Members of the class must be
-- capable of representing fragments (strings that are meant to be
-- interpreted in the language) and text (strings that are just text
-- in which each character stands for itself).  Further, members of
-- the class must provide an observer method 'natrep' that gets the
-- interpretable-string representation of an Language value.  If this
-- string contains any literal text, the text must be properly escaped
-- so that it cannot be accidentally interpreted as fragments (hence
-- the name of the class).

class Language l where
    litfrag  :: String -> l -- | Represents a fragment of the language
    littext  :: String -> l -- | Represents literal text
    natrep   :: l -> String -- | Gets the native-language representation
    language :: l -> String -- | Name of the language


-- | A SafeString is a type-safe string container.  You can construct
-- a SafeString for any Language type, and the type system will
-- prevent you from accidentally mixing the SafeString with plain old
-- strings and incompatible SafeStrings.

data Language l => SafeString l
    = SSEmpty
    | SSText (forall l2 . Language l2 => SafeString l2)
    | SSFragment l
    | SSCat (SafeString l) (SafeString l)


-- The following functions are the only way a plain-old string may
-- gain admittance to a SafeString type.  By calling one of these
-- functions, the programer certifies that the string represents
-- either a literal language fragment or literal text and gets
-- bad a SafeString value of the desired language.

frag,text :: Language l => String -> SafeString l
frag f     = SSFragment (litfrag f)
text s     = SSFragment (littext s)


-- The function 'polytext' certifies that a string represents text and
-- returns a polymorphic SafeString that can be used with any language
-- (which is safe because all languages admit text).

polytext :: String -> SafeString l
polytext s = SSText (text s)


-- | Creates an empty SafeString.

empty     :: Language l => SafeString l
empty      = SSEmpty


-- The following functions can be used to combine SafeStrings of
-- the same language.

cat       :: Language l => [SafeString l] -> SafeString l
cat        = foldr (+++) empty

(+++)     :: Language l => SafeString l -> SafeString l -> SafeString l
(+++)      = SSCat


-- | Tests whether a SafeString represents an empty string (i.e., when
--   rendered).

isEmpty ss   = null (render ss)


-- | Renders a SafeString into a string.

render :: Language l => SafeString l -> String
render ss = renders ss ""

renders SSEmpty        =  id
renders (SSText ss)    =  renders ss
renders (SSFragment a) =  (natrep a ++)
renders (SSCat l r)    =  renders l . renders r


-- | Show instance for Language types.  This lets us see
--   the language and content of safe strings.

instance Language l => Show (SafeString l) where
    showsPrec _ ss =
        (lang ss ++) . (":\"" ++) . renders ss . ('"':)


-- | Returns the name of the language the safe string represents.
--   Note that only the type of 'e', and not its value, is used.

lang ss =
    let SSFragment e = ss in language (undefined `asTypeOf` e)

