module Text.Safe.Strings.TH ( q, declareSafeString ) where

-- Tom Moertel <tom@moertel.com>
-- 2006-03-24

import Text.Safe.Strings.Core

import Control.Monad (liftM)
import Language.Haskell.TH
import qualified Text.ParserCombinators.ReadP as P

-- | AST for things that can take part in template
--   interpolation.  See 'q' and 'xparse'.

data SpecPart
    = SFrag String  -- ^ language fragment
    | SIFrag String -- ^ insert fragment by variable reference
    | SIShow String -- ^ insert rendered variable via show
    | SITxt String  -- ^ insert literal text variable
    | SIRTxt String -- ^ insert rendered safe string var as text
  deriving Show


{- | Compile-time-type-checked safe template interpolation.

Examples:

  $(q "fragment") constructs a safe string, in a given language,
  from a literal-fragment template:

      $(q "My favorite food is <em>a PB&amp;J sandwich</em>.") :: Xml

  #{var} interpolates the contents of 'var' into the expanded
  template.  'var' must be of type String.  Its contents are taken to
  be literal text and will be escaped to prevent conflicts with the
  resulting safe string's language.

      let var = "a PB&J sandwich" in
      $(q "My favorite food is <em>#{var}</em>.") :: Xml

      ==> xml:"My favorite food is <em>a PB&amp;J sandwich</em>."

  #{= var} inserts (verbatim) the SafeString in 'var' into the newly
  constructed safe string.  'var' must contain a string in the
  same language as the safe string being constructed.

      let var = frag "a PB&amp;J sandwich" in
      $(q "<em>My favorite food is #{= var}</em>.") :: Xml

      ==> xml:"<em>My favorite food is a PB&amp;J sandwich</em>."

  #{r var} renders the SafeString in 'var' into text and then inserts
  the literal text, escaping as necessary, into the language of the
  newly constructed safe string:

      let var = frag "a PB&amp;J sandwich" in
      $(q "Write <em>#{= var}</em> in XML as <code>#{r var}</code>.") :: Xml

      ==> xml:"Write <em>a PB&amp;J sandwich</em> in XML as
               <code>a PB&amp;amp;J sandwich</code>."

-}

q :: String -> ExpQ
q spec =
    [| cat $(parts) |]
  where
    parts = case xparse spec of
        []   -> error ("bad template specification: " ++ show spec)
        ps:_ -> foldr gen [| [] |] ps
    gen p ps' = (\p' -> [| $p' : $ps' |]) $ case p of
        SFrag s  -> [| frag $(litE (stringL s))         |]
        SIFrag s -> [| $(varE (mkName s))               |]
        SIShow s -> [| text (show $(varE (mkName s)))   |]
        SITxt s  -> [| text $(varE (mkName s))          |]
        SIRTxt s -> [| text (render $(varE (mkName s))) |]

{-
q spec =
    [| cat $(parts) |]
  where
    parts = case xparse spec of
        []   -> error ("bad template specification: " ++ show spec)
        ps:_ -> gen ps
    gen ps = case ps of
        []           -> [| [] |]
        SFrag s : r  -> [| frag $(litE (stringL s))         : $(gen r) |]
        SIFrag s : r -> [| $(varE (mkName s))               : $(gen r) |]
        SIShow s : r -> [| text (show $(varE (mkName s)))   : $(gen r) |]
        SITxt s : r  -> [| text $(varE (mkName s))          : $(gen r) |]
        SIRTxt s : r -> [| text (render $(varE (mkName s))) : $(gen r) |]
-}

-- | Parse a string into literal language fragments and/or embedded
-- variable references.  References may be to literal text, fragments
-- in the same language, and fragments of arbitrary languages that are
-- to be rendered and then inserted as literal text.
--
-- Example: "fragment #{litref} #{=fragref} #{r renderThisFragRef}"


xparse :: String -> [[SpecPart]]
xparse spec = do
    (result, "") <- P.readP_to_S templateP spec
    return result
 where
    templateP = do
        P.many ((liftM SFrag (P.munch1 (/= '#'))) P.<++
                interpolationP P.<++
                liftM SFrag (P.string "#"))

    interpolationP = do
        P.string "#{"
        spec <- P.manyTill P.get (P.char '}')
        return $ case spec of
          'r':' ':var -> SIRTxt (strip var)
          's':' ':var -> SIShow (strip var)
          '=':var     -> SIFrag (strip var)
          var         -> SITxt  (strip var)

strip = frontAndBack (dropWhile (== ' '))
frontAndBack f = reverse . f . reverse . f


-- | Declare a new kind of safe string for a named language 'lang',
-- where 'typ' is the name of the type you want to declare for
-- safe strings of that language and 'escapeExpQ' is a Template-
-- Haskell quoted expression that gives the escaping function
-- for that language.  See SafeSql.hs and SafeUrl.hs for
-- examples of how to use this function.

declareSafeString lang typ escapeExpQ = do
    let wrapper =  typ ++ "SafeStringWrapper"
    let langN   =  mkName lang
    let typeN   =  mkName typ
    let renderN =  mkName ("render" ++ typ)
    wrapperN    <- newName wrapper
    unWrapperN  <- newName ("un" ++ wrapper)
    escapeExp   <- escapeExpQ
    return $
        [ -- type Xml = SafeString XmlString
          TySynD typeN [] (AppT (ConT ''SafeString) (ConT wrapperN))

          -- newtype XmlString = XmlString { unXmlString :: String }
          --   deriving Show
        , NewtypeD [] wrapperN []
              (RecC wrapperN [(unWrapperN,NotStrict,ConT ''String)]) [''Show]


          -- instance Language XmlString where ...
        , InstanceD [] (AppT (ConT ''Language) (ConT wrapperN))
              [ ValD (VarP 'litfrag) (NormalB (ConE wrapperN)) []
              , ValD (VarP 'littext)
                     (NormalB (InfixE (Just (ConE wrapperN))
                                      (VarE ('(.)))
                                      (Just escapeExp))) []
              , ValD (VarP 'natrep) (NormalB (VarE unWrapperN)) []
              , ValD (VarP 'language) (NormalB (AppE (VarE 'const)
                                                    (LitE (StringL lang)))) []]

          -- xml :: String -> Xml
          -- xml = frag
        , SigD langN (AppT (AppT ArrowT (ConT ''String)) (ConT typeN))
        , ValD (VarP langN) (NormalB (VarE 'frag)) []

          -- renderXml :: Xml -> String
          -- renderXml = render
        , SigD renderN (AppT (AppT ArrowT (ConT typeN)) (ConT ''String))
        , ValD (VarP renderN) (NormalB (VarE 'render)) []]
